open Printf

type program = Program of string * vardec list * istmt list 
and vardec = info * string * ttype
and istmt = info * stmt
and stmt =  
    Assign of string * iexpr
  | Read of string
  | Write of iexpr
  | Sequence of istmt list
  | IfThenElse of iexpr * istmt * istmt option
  | While of iexpr * istmt
  | For of string * iexpr * iexpr * int * istmt
and ttype = 
    TInteger
  | TReal
  | TBoolean
  | TUndef
  | TError
and iexpr = info * expr
and expr = 
    CInteger of int
  | CReal of float
  | Time
  | CBoolean of bool
  | Var of string 
  | Par of iexpr
  | Op of ioperator * iexpr list
and ioperator = info * operator
and operator =
  | OAdd | OSub | OMul | OFrac | ODiv | OMod  
  | OPos | ONeg 
  | ONot | OAnd | OOr  
  | OEq  | ONeq | OLess | OLeq | OGrea | OGeq
and loc = { line: int; col: int }
and info = { locL: loc; locR: loc; typ: ttype }

let cinfo (a,b,t) = { locL=a.locL; locR=b.locR; typ=t }

let setType t r = { locL=r.locL; locR=r.locR; typ=t }

let gI (inf,_) = inf
let gE (_,e) = e

let symtab = Hashtbl.create 10007 

let curr_address_Integer = ref 0
let curr_address_Real    = ref 0
let curr_address_Boolean = ref 0

let add_vartype ((s:string),(t:ttype)) = 
  match t with
  | TInteger -> 
      Hashtbl.add symtab s (t, !curr_address_Integer);
      incr curr_address_Integer
  | TReal ->
      Hashtbl.add symtab s (t, !curr_address_Real);
      incr curr_address_Real
  | TBoolean ->
      Hashtbl.add symtab s (t, !curr_address_Boolean);
      incr curr_address_Boolean
  
let errMessages = ref [] 

let print_Loc loc = sprintf "(%d,%d)" loc.line loc.col

let print_Range inf = print_Loc inf.locL ^ "-" ^ print_Loc inf.locR

let addErrMsg inf msg = 
  errMessages := !errMessages @ [ print_Range inf ^ ": " ^ msg ] 
                                         
 
let lookup_type (info,name) = 
  try let (t,_) = Hashtbl.find symtab name in t 
  with Not_found -> addErrMsg info ("name " ^ name ^ " not found"); TUndef 

let lookup_address (info,name) = 
  try let (_,a) = Hashtbl.find symtab name in a 
  with Not_found -> addErrMsg info ("name " ^ name ^ " not found"); -1 
