%{
open Defs
open Llexer
open Typecheck

let parse_error s = addErrMsg (mr (!curr_line,!curr_col,String.length !curr_tok)) 
                              ("unspecified parse error at token " ^ !curr_tok)
%}

%token <Defs.info * int>TINTEGER      /* integer literals */
%token <Defs.info * float>TREAL       /* real literals */
%token <Defs.info * bool>TBOOLEAN     /* boolean literals */
%token <Defs.info * string>TIDENT     /* variables */
%token <Defs.info * Defs.ttype>TTYPECONST
%token <Defs.info>TPLUS 
%token <Defs.info>TMINUS 
%token <Defs.info>TTIMES 
%token <Defs.info>TSLASH
%token <Defs.info>TLPAREN 
%token <Defs.info>TRPAREN
%token <Defs.info>TSEMICOLON 
%token <Defs.info>TCOMMA 
%token <Defs.info>TCOLON 
%token <Defs.info>TDOT
%token <Defs.info>TASSIGN 
%token <Defs.info>TFOR
%token <Defs.info>TTO
%token <Defs.info>TDOWNTO
%token <Defs.info>TDO
%token <Defs.info>TWHILE
%token <Defs.info>TIF
%token <Defs.info>TTHEN
%token <Defs.info>TELSE
%token <Defs.info>TNOT
%token <Defs.info>TAND
%token <Defs.info>TOR
%token <Defs.info>TDIV
%token <Defs.info>TMOD                
%token <Defs.info>TREAD 
%token <Defs.info>TTIME
%token <Defs.info>TWRITE
%token <Defs.info>TPROGRAM
%token <Defs.info>TVAR
%token <Defs.info>TBEGIN 
%token <Defs.info>TEND 
%token <Defs.info>TEQ
%token <Defs.info>TNEQ
%token <Defs.info>TLESS
%token <Defs.info>TLEQ
%token <Defs.info>TGREA
%token <Defs.info>TGEQ
%nonassoc TEQ TNEQ TLESS TLEQ TGREA TGEQ
%left TPLUS TMINUS TOR       
%left TTIMES TSLASH TDIV TMOD TAND    
%nonassoc TNOT     
%start main             /* the entry point */
%type <Defs.program> main
%type <(Defs.info * string) list> idlist
%type <Defs.vardec list> vardec
%type <Defs.vardec list> vardecs
%type <Defs.vardec list * Defs.istmt list> blockStep1
%type <Defs.istmt list> blockStep2
%type <Defs.istmt list> sequence
%type <Defs.istmt> stmt
%type <Defs.iexpr> exp
%%
main:
  TPROGRAM TIDENT TSEMICOLON blockStep1 TDOT { Program (gE $2,gI $4, gE $4) }
  ;
idlist:
  TIDENT                 { [$1] }
| idlist TCOMMA TIDENT { $3::$1 } 
;
vardec:
  idlist TCOLON TTYPECONST { let ids = List.rev $1 in
                           List.iter (fun (_,s) -> add_vartype (s,gE $3)) ids;
                           List.map (fun (r,v) -> (r,v,gE $3)) $1             }
  ;
vardecs:
  vardec                { $1 }
| vardecs TSEMICOLON vardec { $1 @ $3 }
blockStep1:
  TVAR vardecs TSEMICOLON blockStep2 { ($2,$4) }
| blockStep2 { ([],$1) }    ;
blockStep2:
  TBEGIN sequence TEND { List.rev $2 }
  ; 
sequence:
  stmt                       { [$1] }
| sequence TSEMICOLON stmt    { $3 :: $1 }
    ;
stmt:
  TIDENT TASSIGN exp            { tcAssign ($1,$3) } 
| TREAD TIDENT                  { (cinfo ($1,gI $2,TUndef), Read (gE $2)) }
| TWRITE exp                    { (cinfo ($1,gI $2,TUndef), Write $2) }
| TBEGIN sequence TEND          { (cinfo ($1,$3,TUndef), Sequence (List.rev $2)) }
| TIF exp TTHEN stmt            { tcIf ($1,$2,$4,None) }
| TIF exp TTHEN stmt TELSE stmt { tcIf ($1,$2,$4,Some $6) } 
| TWHILE exp TDO stmt           { tcWhile ($1,$2,$4) }
| TFOR TIDENT TASSIGN exp TTO exp TDO stmt     { tcFor ($1,$2,$4,$6,1,$8) }
| TFOR TIDENT TASSIGN exp TDOWNTO exp TDO stmt { tcFor ($1,$2,$4,$6,-1,$8) }
;
exp:
  TINTEGER                { (setType TInteger (gI $1), CInteger (gE $1)) }
| TREAL                   { (setType TReal    (gI $1), CReal    (gE $1)) }
| TTIME                   { (cinfo ($1,$1,TReal), Time) }
| TBOOLEAN                { (setType TBoolean (gI $1), CBoolean (gE $1)) }
| TIDENT                  { (setType (lookup_type $1) (gI $1), Var (gE $1)) } 
| TLPAREN exp  TRPAREN    { (cinfo ($1,$3,(gI $2).typ), Par $2) }
| exp  TPLUS exp           { tcBinAriOp (OAdd,$2,$1,$3) }
| exp  TMINUS exp          { tcBinAriOp (OSub,$2,$1,$3) }
| exp  TTIMES exp          { tcBinAriOp (OMul,$2,$1,$3) }
| exp  TSLASH exp          { tcBinAriOp (OFrac,$2,$1,$3) }
| exp  TDIV exp            { tcBinAriOp (ODiv,$2,$1,$3) }
| exp  TMOD exp            { tcBinAriOp (OMod,$2,$1,$3) }
| TNOT exp                 { tcUnOp (ONot,$1,$2) }
| exp  TAND exp            { tcBinBoolOp (OAnd,$2,$1,$3) }
| exp  TOR  exp            { tcBinBoolOp (OOr,$2,$1,$3) }
| exp  TEQ exp             { tcCmpOp (OEq,$2,$1,$3) }
| exp  TNEQ exp            { tcCmpOp (ONeq,$2,$1,$3) }
| exp  TLESS exp           { tcCmpOp (OLess,$2,$1,$3) }
| exp  TLEQ exp            { tcCmpOp (OLeq,$2,$1,$3) }
| exp  TGREA exp           { tcCmpOp (OGrea,$2,$1,$3) }
| exp  TGEQ exp            { tcCmpOp (OGeq,$2,$1,$3) }
| TPLUS  exp %prec TNOT  { tcUnOp (OPos,$1,$2) }
| TMINUS exp %prec TNOT  { tcUnOp (ONeg,$1,$2)   }
    ;
  
