open Printf
open Defs 
open Pprint


let maxType = function
    (TInteger,TReal) -> TReal
  | (TReal,TInteger) -> TReal
  | (a,b) -> if a=b then a else TError

let tcUnOp (op,rngOp,exp) = 
  let inf2 = gI exp in
  let baseType = match op with
                  OPos | ONeg -> TInteger
                | ONot -> TBoolean in
  let nt = maxType (baseType, inf2.typ) in
  let ninf = { locL=rngOp.locL; locR=inf2.locR; typ=nt } in
  (ninf, Op((rngOp,op),[exp]))
 
let tcBinAriOp (op,rngOp,exp1,exp2) = 
  let inf1 = gI exp1 in
  let inf2 = gI exp2 in
  let nt = 
    match op with
    | OFrac -> maxType (TReal, maxType (inf1.typ, inf2.typ))
    | _     -> maxType (TInteger, maxType (inf1.typ,inf2.typ)) in
  let ninf = { locL=inf1.locL; locR=inf2.locR; typ=nt } in
  (ninf, Op ((rngOp,op),[exp1; exp2]))
  
let tcBinBoolOp (op,rngOp,exp1,exp2) = 
  let inf1 = gI exp1 in
  let inf2 = gI exp2 in
  let nt =  maxType (TBoolean, maxType (inf1.typ, inf2.typ)) in 
  let ninf = { locL=inf1.locL; locR=inf2.locR; typ=nt } in
  (ninf, Op ((rngOp,op),[exp1;exp2]))
 
let tcCmpOp (op,rngOp,exp1,exp2) = 
  let inf1 = gI exp1 in
  let inf2 = gI exp2 in
  let _ = match maxType (inf1.typ,inf2.typ)  with
    TUndef | TError -> addErrMsg rngOp "Type error in comparison"
  | _ -> () in
  let ninf = { locL=inf1.locL; locR=inf2.locR; typ=TBoolean } in
  (ninf, Op ((rngOp,op),[exp1;exp2]))


let tcAssign (name,exp) = 
   let declared_type = lookup_type name in
   let rType = maxType (declared_type,(gI exp).typ) in
   if not (rType = declared_type)
   then 
     if declared_type = TUndef
     then addErrMsg (gI name) (sprintf "variable %s is undeclared" (gE name))
     else 
       begin 
	 let s1 = sprintf "variable %s declared to store " (gE name) in
	 let s2 = pprint_Type declared_type in
         let s3 = " but assigned expression of type " in 
         let s4 = pprint_Type ((gI exp).typ) in
         addErrMsg (gI exp) (s1^s2^s3^s4)
       end;
   let ninf = cinfo (gI name, gI exp,rType) in
   (ninf, Assign (gE name,exp))
 

let tcIf (ifsym,cond,thenB,maybeElseB) =
  if (gI cond).typ <> TBoolean then addErrMsg (gI cond) "type error in if condition";
  let right = match maybeElseB with
                None -> thenB
              | Some x -> x in
  let inf = cinfo (ifsym,gI right,TUndef) in
  (inf, IfThenElse (cond,thenB,maybeElseB))

let tcWhile (whilesym,cond,body) = 
  if (gI cond).typ <> TBoolean then addErrMsg (gI cond) "type error in while condition";
  let inf = cinfo (whilesym,gI body,TUndef) in
  (inf, While (cond,body))

let tcFor (forsym,(var_inf,var_s),estart,eend,step,body) =
  let _ = match lookup_type (var_inf,var_s) with
    TInteger -> ()
  | TReal -> addErrMsg var_inf "loop variable must have type integer, but has type real"
  | TBoolean -> addErrMsg var_inf "Loop variable must have type integer, but has type boolean"
  | TUndef -> addErrMsg var_inf "loop variable not declared"
  | TError -> addErrMsg var_inf "internal error: type error with loop variable" in
  if (gI estart).typ <> TInteger 
     then addErrMsg (gI estart) "first loop bound must be of type integer";
  if (gI eend).typ <> TInteger 
     then addErrMsg (gI eend) "second loop bound must be of type integer";
  let inf = cinfo (forsym,gI body,TUndef) in
  (inf, For (var_s,estart,eend,step,body))
